#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <glib.h>
#include <unistd.h>

#include "common.h"

static int
destroy_tcp_connection(tcp_con_t *con) {
  if (!con)
    return -EINVAL;

  if (con->self)
    freeaddrinfo(con->self);

  close(con->soc);

  return 0;
}

static int
parse_request(request_msg_t *req,const char *request_string){
  size_t len;
  long int_val;
  char *sp=NULL;
  char *ep=NULL;
  char *buffer;
  ssize_t remains;
  int rc=0;

  if  ( (!request_string) || (!req) )
    return -EINVAL;

  buffer=strdup(request_string);
  if (!buffer)
    return -ENOMEM;
  len=strlen(request_string);
  _assert(len>0);
  remains=len;
  /*
   * パケット番号
   */
  sp=buffer;
  ep=memchr(sp, ':', remains);
  if (!ep) {
    rc=-EINVAL;
    goto err_out;
  }

  *ep='\0';
  remains =len - ((unsigned long)ep-(unsigned long)buffer);
  if (remains<=0) {
    rc=-EINVAL;
    goto err_out;
  }
  ++ep;
  int_val=strtol(sp, (char **)NULL, 16);
  req->pkt_no=int_val;
  dbg_out("pktno:%ld\n",req->pkt_no);
  sp=ep;

  /*
   * ファイルID
   */
  ep=memchr(sp, ':', remains);
  if (!ep) {
    rc=-EINVAL;
    goto err_out;
  }
  *ep='\0';
  remains =len - ((unsigned long)ep-(unsigned long)buffer);
  if (remains<=0) {
    rc=-EINVAL;
    goto err_out;
  }
  ++ep;
  int_val=strtol(sp, (char **)NULL, 16);
  req->fileid=int_val;
  dbg_out("fileid:%d(%x)\n",req->fileid,req->fileid);
  sp=ep;

  /*
   * オフセット
   */
  int_val=strtol(sp, (char **)NULL, 16);
  req->offset=int_val;
  dbg_out("offset:%d(%x)\n",req->offset,req->offset);
err_out:
  free(buffer);
  return rc;
}
static int
tcp_transfer_file(tcp_con_t *con,const char *path,const size_t size,off_t offset){
  int fd;
  int rc;
  char buff[TCP_FILE_BUFSIZ];
  char dummy[1];
  ssize_t read_len;
  ssize_t file_remains;
  ssize_t soc_remains;
  size_t write_len;
  size_t total_write;
  char *wp;

  if ( (!con) || (!path) )
    return -EINVAL;

  fd=open(path,O_RDONLY);
  if (fd<0)
    return -errno;

  rc=lseek(fd, offset, SEEK_SET);
  if (rc<0) {
    rc=-errno;
    goto close_out;
  }
  total_write=0;
  file_remains=size;

  while(file_remains>0) {
    read_len=read(fd,buff,TCP_FILE_BUFSIZ);
    if (read_len<0) {
      rc=-errno;
      goto close_out;
    }
    file_remains -= read_len;
    soc_remains=read_len;
    wp=buff;
    if (wait_socket(con->soc,WAIT_FOR_WRITE,TCP_SELECT_SEC)<0) {
      err_out("Can not send socket\n");
      goto close_out;
    }
    while(soc_remains>0) {
      write_len=send(con->soc,wp,soc_remains,MSG_NOSIGNAL);
      if (write_len<0) {
	if (errno==EINTR)
	  continue;
	goto close_out;
      }
      wp += write_len;
      total_write += write_len;
      soc_remains -= write_len;
    }
    dbg_out("transfer %s %d/%d(%d)\n",path,total_write,size,file_remains);
  }
  dbg_out("Wait for closing by peer.\n");
  sock_set_recv_timeout(con->soc,TCP_CLOSE_WAIT_SEC);
  if (wait_socket(con->soc,WAIT_FOR_READ,TCP_SELECT_SEC)<0) {
    dbg_out("OK timeout select :%s(%d)\n",strerror(errno),errno);
  }

  memset(dummy, 0, sizeof(dummy));
  rc=recv(con->soc, dummy,sizeof(dummy),0); /* 相手のクローズ検出 */
  if (rc<0)
    dbg_out("OK timeout:%s(%d)\n",strerror(errno),errno);
  rc=0;
 close_out:
  close(fd);
  if (rc<0)
    dbg_out("Can not send file:%s %d\n",strerror(errno),errno);
  return rc;
}
int
tcp_enable_keepalive(const tcp_con_t *con){
  int rc;
  int flag;

  if (!con) 
    return -EINVAL;

  flag=1;
  rc=setsockopt(con->soc, SOL_SOCKET, SO_KEEPALIVE, (void*)&flag, sizeof(int));
  if (rc<0) {
    err_out("Can not set keepalive:%s(%d)\n",strerror(errno),errno);
    return -errno;
  }

  return 0;
}
int
tcp_setup_client(int family,const char *ipaddr,int port,tcp_con_t *con){
  int rc;
  int sock=-1;
  struct sockaddr_in server;
  struct hostent *host;
  struct addrinfo *info=NULL;

  if ( (!ipaddr) || (!con) )
    return -EINVAL;

  rc=setup_addr_info(&info, ipaddr,port,SOCK_STREAM,family);
  if (rc<0)
    return rc;

  sock=socket(info->ai_family, info->ai_socktype,info->ai_protocol);

  if (sock<0) {
    err_out("Can not create socket: %s(errno:%d)\n",strerror(errno),errno);
    rc=-errno;
    goto error_out;    
  }

  if (connect(sock, info->ai_addr, info->ai_addrlen)<0){
    err_out("Can not connect addr:%s port:%d %s (errno:%d)\n",
	    ipaddr,
	    port,
	    strerror(errno),
	    errno);
    rc=-errno;
    goto error_out;    
  }
   
  con->soc=sock;
  con->family=info->ai_family;
  con->self=info;

  return 0;
 error_out:
  freeaddrinfo(info);
  return rc;
}
int
tcp_set_buffer(tcp_con_t *con) {
  if (!con)
    return -EINVAL;

 return sock_set_buffer(con->soc);
}
gpointer
ipmsg_tcp_recv_thread(gpointer data){
  ssize_t recv_len;
  size_t addr_len;
  int rc;
  tcp_con_t *con;
  char recv_buf[_MSG_BUF_SIZE];
  int count=200;
  struct addrinfo *info;

  con=(tcp_con_t *)data;
  if (!con) {
    err_out("No connection recived\n");
    g_assert(con);
    return NULL;
  }
  rc=tcp_enable_keepalive(con);
  if (rc<0)
    return NULL;
  
#if 0
  rc=tcp_set_buffer(con);
  if (rc<0)
    return NULL;
#endif
  while(1) {
    if (wait_socket(con->soc,WAIT_FOR_READ,TCP_SELECT_SEC)<0) {
      err_out("Can not send socket\n");
      return NULL;
    }else{
      memset(recv_buf, 0, sizeof(recv_buf));
      info=con->self;
      recv_len=recv(con->soc,
		    recv_buf,
		    _MSG_BUF_SIZE,
		    MSG_PEEK);

      if (recv_len<=0) {
	if (errno)
	  err_out("Can not peek message %s(errno:%d)\n",strerror(errno),errno);
	destroy_tcp_connection(con);
	free(con);
	return NULL;
      }

      dbg_out("tcp peek read:%d %s\n",recv_len,recv_buf);

      recv_len=recv(con->soc,
		    recv_buf,
		    recv_len,
		    MSG_PEEK);


      dbg_out("tcp read:%d %s\n",recv_len,recv_buf);
      if (recv_len>0) {
	msg_data_t msg;
	request_msg_t req;
	char *path;
	size_t size;

	recv_buf[recv_len-1]='\0';
	memset(&req,0,sizeof(request_msg_t));

	init_message_data(&msg);
	parse_message(&msg,recv_buf,recv_len);
	parse_request(&req,msg.message);

	if (refer_attach_file(req.pkt_no,req.fileid,(const char **)&path,&size)<0) {
	  err_out("Can not find message:pktno %ld id : %d\n",
		  req.pkt_no,
		  req.fileid);
	}else{
	  dbg_out("transfer:%s (%d)\n",path,size);
	  tcp_transfer_file(con,path,size,0);
	  release_message_data(&msg);
	  break;
	}
	release_message_data(&msg);
	break;
      }
    }
    --count;
    if (!count)
      break;
  }
  if (con->self) /* closeは相手側で行うので, destroy_tcp_connectionは呼び出せない
		  * (Win版ipmsgの仕様).  
		  */
    freeaddrinfo(con->self); 
  free(con);

  return NULL;
}
gpointer 
ipmsg_tcp_server_thread(gpointer data){
  int rc;
  size_t len;
  int sock;
  int tcp_socket;
  int port;
  int reuse;
  int family;
  tcp_con_t *client_con=NULL;
  struct addrinfo *info=NULL;

  family=(int)data;
  rc=setup_addr_info(&info, 
		     NULL,
		     hostinfo_refer_ipmsg_port(),
		     SOCK_STREAM,family);
  if (rc<0)
    return NULL;

  tcp_socket = socket(info->ai_family,
		      info->ai_socktype,
		      info->ai_protocol);

  if (tcp_socket < 0) {
    err_out("Can not create socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  reuse=1;
  if (setsockopt(tcp_socket,SOL_SOCKET, SO_REUSEADDR, (const char *)&reuse, sizeof(reuse))<0) {
    err_out("Can not set sock opt:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  if (bind(tcp_socket, info->ai_addr, info->ai_addrlen) != 0) {
    err_out("Can not bind socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  if (listen(tcp_socket, TCP_LISTEN_LEN) != 0) {
    err_out("Can not listen socket:%s (%d)\n",strerror(errno),errno);
    goto error_out;
  }

  while (1) {
    GThread *handler;
    struct addrinfo *client_info=NULL;

    client_con=malloc(sizeof(tcp_con_t));
    g_assert(client_con);
    dbg_out("accept\n");

    
    rc=setup_addr_info(&client_info, 
		     NULL,
		     hostinfo_refer_ipmsg_port(),
		     SOCK_STREAM,family);
    
    if (rc<0)
      goto free_client_out;
    
    sock = accept(tcp_socket, client_info->ai_addr,&(client_info->ai_addrlen));
    if (sock < 0) {
      err_out("Can not accept:%s (%d)\n",strerror(errno),errno);
      goto free_client_out;
    }

    client_con->soc=sock;
    client_con->self=client_info;
    handler=g_thread_create(ipmsg_tcp_recv_thread,
                            client_con,
			    FALSE,
			    NULL);
    if (!handler) {
      err_out("Can not create handler thread.\n");
      goto free_client_out;
    }
  } 
  return NULL;

 free_client_out:  
  destroy_tcp_connection(client_con);
 error_out:
  freeaddrinfo(info);
  return NULL;
}
